/* Egoboo - graphicmad.c
* Character model drawing code.
*/

/*
This file is part of Egoboo.

Egoboo is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Egoboo is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "egoboo.h"
#include "Md2.h"
#include "id_md2.h"
#include "Log.h"
#include <SDL_opengl.h>

/* Storage for blended vertices */
static GLfloat md2_blendedVertices[MD2_MAX_VERTICES][3];
static GLfloat md2_blendedNormals[MD2_MAX_VERTICES][3];

/* blend_md2_vertices
* Blends the vertices and normals between 2 frames of a md2 model for animation.
*
* NOTE: Only meant to be called from draw_textured_md2, which does the necessary
* checks to make sure that the inputs are valid.  So this function itself assumes
* that they are valid.  User beware!
*/
static void blend_md2_vertices(Uint8 character, const Md2Model *model, int from_, int to_, float lerp)
{
  Md2Frame *from, *to;
  int numVertices, i;
  Uint8 ambilevel     = chrlightambi[character];
  Uint8 speklevel     = chrlightspek[character];
  Uint8 sheen         = chrsheen[character];
  Uint8 lightrotation = (chrturnleftright[character] + chrlightturnleftright[character]) >> 8;
  Uint16 lightold, lightnew;
  float spekularity = ((float)sheen / (float)MAXSPEKLEVEL);

  from = &model->frames[from_];
  to = &model->frames[to_];
  numVertices = model->numVertices;


  if (lerp <= 0)
  {
    // copy the vertices in frame 'from' over
    for (i = 0;i < numVertices;i++)
    {
      md2_blendedVertices[i][0] = from->vertices[i].x;
      md2_blendedVertices[i][1] = from->vertices[i].y;
      md2_blendedVertices[i][2] = from->vertices[i].z;

      md2_blendedNormals[i][0] = kMd2Normals[from->vertices[i].normal][0];
      md2_blendedNormals[i][1] = kMd2Normals[from->vertices[i].normal][1];
      md2_blendedNormals[i][2] = kMd2Normals[from->vertices[i].normal][2];

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[from_][i]] + spekularity * 255 * spek_global[lightrotation][madvrta[from_][i]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][i];
      chrvrta[character][i] = 0.9 * lightold + 0.1 * lightnew;
    }
  }
  else if (lerp >= 1.0f)
  {
    // copy the vertices in frame 'to'
    for (i = 0;i < numVertices;i++)
    {
      md2_blendedVertices[i][0] = to->vertices[i].x;
      md2_blendedVertices[i][1] = to->vertices[i].y;
      md2_blendedVertices[i][2] = to->vertices[i].z;

      md2_blendedNormals[i][0] = kMd2Normals[to->vertices[i].normal][0];
      md2_blendedNormals[i][1] = kMd2Normals[to->vertices[i].normal][1];
      md2_blendedNormals[i][2] = kMd2Normals[to->vertices[i].normal][2];

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[to_][i]] + spekularity * 255 * spek_global[lightrotation][madvrta[to_][i]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][i];
      chrvrta[character][i] = 0.9 * lightold + 0.1 * lightnew;
    }
  }
  else
  {
    // mix the vertices
    for (i = 0;i < numVertices;i++)
    {
      md2_blendedVertices[i][0] = from->vertices[i].x +
                                  (to->vertices[i].x - from->vertices[i].x) * lerp;
      md2_blendedVertices[i][1] = from->vertices[i].y +
                                  (to->vertices[i].y - from->vertices[i].y) * lerp;
      md2_blendedVertices[i][2] = from->vertices[i].z +
                                  (to->vertices[i].z - from->vertices[i].z) * lerp;

      md2_blendedNormals[i][0] = kMd2Normals[from->vertices[i].normal][0] +
                                 (kMd2Normals[to->vertices[i].normal][0] - kMd2Normals[from->vertices[i].normal][0]) * lerp;
      md2_blendedNormals[i][0] = kMd2Normals[from->vertices[i].normal][1] +
                                 (kMd2Normals[to->vertices[i].normal][1] - kMd2Normals[from->vertices[i].normal][1]) * lerp;
      md2_blendedNormals[i][0] = kMd2Normals[from->vertices[i].normal][2] +
                                 (kMd2Normals[to->vertices[i].normal][2] - kMd2Normals[from->vertices[i].normal][2]) * lerp;


      lightold = MIN(255, speklevel * spek_local[lightrotation][madvrta[from_][i]] + spekularity * 255 * spek_global[lightrotation][madvrta[from_][i]]);
      lightold = lighttospek[sheen][lightold] + ambilevel + (lightambi * 255);

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[to_][i]] + spekularity * 255 * spek_global[lightrotation][madvrta[to_][i]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);

      lightnew = (1 - lerp) * lightold + lerp * lightnew;
      lightold = chrvrta[character][i];
      chrvrta[character][i] = 0.9 * lightold + 0.1 * lightnew;
    }
  }
}

/* draw_textured_md2
* Draws a Md2Model in the new format
*/
void draw_textured_md2(Uint16 character, const Md2Model *model, int from_, int to_, float lerp)
{
  int i, numTriangles;
  const Md2TexCoord *tc;
  const Md2Triangle *triangles;
  const Md2Triangle *tri;

  if (model == NULL) return;
  if (from_ < 0 || from_ >= model->numFrames) return;
  if (to_ < 0 || to_ >= model->numFrames) return;

  blend_md2_vertices(character, model, from_, to_, lerp);

  numTriangles = model->numTriangles;
  tc = model->texCoords;
  triangles = model->triangles;

  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);

  glVertexPointer(3, GL_FLOAT, 0, md2_blendedVertices);
  glNormalPointer(GL_FLOAT, 0, md2_blendedNormals);

  glBegin(GL_TRIANGLES);
  {
    for (i = 0;i < numTriangles;i++)
    {
      tri = &triangles[i];

      glTexCoord2fv((const GLfloat*)&(tc[tri->texCoordIndices[0]]));
      glArrayElement(tri->vertexIndices[0]);

      glTexCoord2fv((const GLfloat*)&(tc[tri->texCoordIndices[1]]));
      glArrayElement(tri->vertexIndices[1]);

      glTexCoord2fv((const GLfloat*)&(tc[tri->texCoordIndices[2]]));
      glArrayElement(tri->vertexIndices[2]);
    }
  }
  glEnd();

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
}

//--------------------------------------------------------------------------------------------
void render_mad_lit(Uint16 character)
{
  // ZZ> This function draws an environment mapped model

  GLVERTEX v[MAXVERTICES];
  Uint16 cnt, tnc, entry;
  Uint16 vertex;
  Sint32 temp;

  Uint8  sheen = chrsheen[character];
  float          spekularity = ((float)sheen / (float)MAXSPEKLEVEL);
  Uint16 model = chrmodel[character];
  Uint16 texture = chrtexture[character];
  Uint16 frame = chrframe[character];
  Uint16 lastframe = chrlastframe[character];
  Uint8 lip = chrlip[character] >> 6;
  Uint8 lightrotation = (chrturnleftright[character] + chrlightturnleftright[character]) >> 8;
  Uint8 ambilevel = chrlightambi[character];
  Uint8 speklevel = chrlightspek[character];


  float uoffset = textureoffset[chruoffset[character] >> 8];
  float voffset = textureoffset[chrvoffset[character] >> 8];
  Uint8 rs = chrredshift[character];
  Uint8 gs = chrgrnshift[character];
  Uint8 bs = chrblushift[character];

  float mat_none[4]     = {0, 0, 0, 0};
  float mat_emission[4] = {0, 0, 0, 0};
  float mat_diffuse[4]  = {0, 0, 0, 0};
  float mat_specular[4] = {0, 0, 0, 0};
  float shininess[1] = {2};

  float ftmp;

  // Original points with linear interpolation ( lip )
  switch (lip)
  {
  case 0:  // 25% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[frame][cnt] + temp >> 2);

      temp = madvrty[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[frame][cnt] + temp >> 2);

      temp = madvrtz[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[frame][cnt] + temp >> 2);
    }
    break;

  case 1:  // 50% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float)(madvrtx[frame][cnt] +
                         madvrtx[lastframe][cnt] >> 1);
      v[cnt].y = (float)(madvrty[frame][cnt] +
                         madvrty[lastframe][cnt] >> 1);
      v[cnt].z = (float)(madvrtz[frame][cnt] +
                         madvrtz[lastframe][cnt] >> 1);
    }
    break;

  case 2:  // 75% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[lastframe][cnt] + temp >> 2);
      temp = madvrty[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[lastframe][cnt] + temp >> 2);
      temp = madvrtz[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[lastframe][cnt] + temp >> 2);
    }
    break;

  case 3:  // 100% this frame...  This is the legible one
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float) madvrtx[frame][cnt];
      v[cnt].y = (float) madvrty[frame][cnt];
      v[cnt].z = (float) madvrtz[frame][cnt];
    }
    break;

  }

  ftmp = ((float)(MAXSPEKLEVEL - sheen) / (float)MAXSPEKLEVEL) * ((float)chralpha[character] / 255.0f);
  mat_diffuse[0] = ftmp / (float)(1 << rs) * (float)chralpha[character] / 255.0f;
  mat_diffuse[1] = ftmp / (float)(1 << gs) * (float)chralpha[character] / 255.0f;
  mat_diffuse[2] = ftmp / (float)(1 << bs) * (float)chralpha[character] / 255.0f;


  ftmp = ((float)sheen / (float)MAXSPEKLEVEL) * ((float)chralpha[character] / 255.0f);
  mat_specular[0] = ftmp / (float)(1 << rs);
  mat_specular[1] = ftmp / (float)(1 << gs);
  mat_specular[2] = ftmp / (float)(1 << bs);

  shininess[0] = sheen + 2;

  if (255 != chrlight[character])
  {
    ftmp = (chrlight[character] / 255.0f);
    mat_emission[0] = ftmp / (float)(1 << rs);
    mat_emission[1] = ftmp / (float)(1 << gs);
    mat_emission[2] = ftmp / (float)(1 << bs);
  }

  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,   mat_none);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,   mat_none);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,  mat_emission);

  ATTRIB_PUSH("render_mad_lit", GL_ENABLE_BIT | GL_TRANSFORM_BIT | GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT);
  {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixf(chrmatrix[character].v);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // Choose texture and matrix
    GLTexture_Bind(&TxTexture[texture], CData.texturefilter);

    // Render each command
    glColor4f(1, 1, 1, 1);
    entry = 0;
    for (cnt = 0; cnt < madcommands[model]; cnt++)
    {
      glBegin(madcommandtype[model][cnt]);

      for (tnc = 0; tnc < madcommandsize[model][cnt]; tnc++)
      {
        vertex = madcommandvrt[model][entry];


        glTexCoord2f(madcommandu[model][entry] + uoffset, madcommandv[model][entry] + voffset);
        glNormal3f(kMd2Normals[madvrta[lastframe][vertex]][0], -kMd2Normals[madvrta[lastframe][vertex]][1], -kMd2Normals[madvrta[lastframe][vertex]][2]);
        glVertex3fv(&v[vertex].x);

        entry++;
      }
      glEnd();
    }

    glPopMatrix();
  }
  ATTRIB_POP("render_mad_lit");

}


//--------------------------------------------------------------------------------------------
void render_enviromad(Uint16 character, Uint8 trans)
{
  // ZZ> This function draws an environment mapped model
  //D3DLVERTEX v[MAXVERTICES];
  //D3DTLVERTEX vt[MAXVERTICES];
  //D3DTLVERTEX vtlist[MAXCOMMANDSIZE];
  GLVERTEX v[MAXVERTICES];
  Uint16 cnt, tnc, entry;
  Uint16 vertex;
  Sint32 temp;
  //float z;
  //Uint8 red, grn, blu;
  //DWORD fogspec;

  Uint8  sheen = chrsheen[character];
  float          spekularity = ((float)sheen / (float)MAXSPEKLEVEL);
  Uint16 model = chrmodel[character];
  Uint16 texture = chrtexture[character];
  Uint16 frame = chrframe[character];
  Uint16 lastframe = chrlastframe[character];
  Uint16 framestt = madframestart[chrmodel[character]];
  Uint8 lip = chrlip[character] >> 6;
  Uint8 lightrotation = (chrturnleftright[character] + chrlightturnleftright[character]) >> 8;
  Uint8 ambilevel = chrlightambi[character];
  Uint8 speklevel = chrlightspek[character];
  float uoffset = textureoffset[chruoffset[character] >> 8] + camturnleftrightone;
  float voffset = textureoffset[chrvoffset[character] >> 8];
  Uint8 rs = chrredshift[character];
  Uint8 gs = chrgrnshift[character];
  Uint8 bs = chrblushift[character];
  Uint16 lightnew, lightold;


  // Original points with linear interpolation ( lip )
  switch (lip)
  {
  case 0:  // 25% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[frame][cnt] + temp >> 2);
      temp = madvrty[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[frame][cnt] + temp >> 2);
      temp = madvrtz[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[frame][cnt] + temp >> 2);


      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);

      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f;
      v[cnt].a = (float)(trans) / 255.0f;
    }
    break;

  case 1:  // 50% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float)(madvrtx[frame][cnt] +
                         madvrtx[lastframe][cnt] >> 1);
      v[cnt].y = (float)(madvrty[frame][cnt] +
                         madvrty[lastframe][cnt] >> 1);
      v[cnt].z = (float)(madvrtz[frame][cnt] +
                         madvrtz[lastframe][cnt] >> 1);

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);
      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f;
      v[cnt].a = (float)(trans) / 255.0f;
    }
    break;

  case 2:  // 75% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[lastframe][cnt] + temp >> 2);
      temp = madvrty[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[lastframe][cnt] + temp >> 2);
      temp = madvrtz[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[lastframe][cnt] + temp >> 2);

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);
      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f;
      v[cnt].a = (float)(trans) / 255.0f;
    }
    break;

  case 3:  // 100% this frame...  This is the legible one
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float) madvrtx[frame][cnt];
      v[cnt].y = (float) madvrty[frame][cnt];
      v[cnt].z = (float) madvrtz[frame][cnt];

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);
      v[cnt].a = (float)(trans) / 255.0f;
      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f;
    }
    break;

  }

  // Do fog...
  /*
  if(fogon && chrlight[character]==255)
  {
  // The full fog value
  alpha = 0xff000000 | (fogred<<16) | (foggrn<<8) | (fogblu);

  for (cnt = 0; cnt < madtransvertices[model]; cnt++)
  {
  // Figure out the z position of the vertex...  Not totally accurate
  z = (v[cnt].z * chrscale[character]) + chrmatrix[character](3,2);

  // Figure out the fog coloring
  if(z < fogtop)
  {
  if(z < fogbottom)
  {
  v[cnt].specular = alpha;
  }
  else
  {
  z = 1.0 - ((z - fogbottom)/fogdistance);  // 0.0 to 1.0...  Amount of fog to keep
  red = fogred * z;
  grn = foggrn * z;
  blu = fogblu * z;
  fogspec = 0xff000000 | (red<<16) | (grn<<8) | (blu);
  v[cnt].specular = fogspec;
  }
  }
  else
  {
  v[cnt].specular = 0;
  }
  }
  }
  else
  {
  for (cnt = 0; cnt < madtransvertices[model]; cnt++)
  v[cnt].specular = 0;
  }
  */

  ATTRIB_PUSH("render_enviromad", GL_TRANSFORM_BIT | GL_ENABLE_BIT | GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
  {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixf(chrmatrix[character].v);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // Choose texture and matrix
    GLTexture_Bind(&TxTexture[texture], CData.texturefilter);

    glEnable(GL_TEXTURE_GEN_S);       // Enable Texture Coord Generation For S (NEW)
    glEnable(GL_TEXTURE_GEN_T);       // Enable Texture Coord Generation For T (NEW)

    // Render each command
    entry = 0;
    for (cnt = 0; cnt < madcommands[model]; cnt++)
    {
      glBegin(madcommandtype[model][cnt]);

      for (tnc = 0; tnc < madcommandsize[model][cnt]; tnc++)
      {
        vertex = madcommandvrt[model][entry];

        glColor4fv(&v[vertex].r);
        glNormal3f(kMd2Normals[madvrta[lastframe][vertex]][0], -kMd2Normals[madvrta[lastframe][vertex]][1], -kMd2Normals[madvrta[lastframe][vertex]][2]);
        glVertex3fv(&v[vertex].x);

        entry++;
      }
      glEnd();
    }

    glPopMatrix();
  }
  ATTRIB_POP("render_enviromad");
}




//--------------------------------------------------------------------------------------------
void render_texmad(Uint16 character, Uint8 trans)
{
  // ZZ> This function draws a model

  GLVERTEX v[MAXVERTICES];
  Uint16 cnt, tnc, entry;
  Uint16 vertex;
  Sint32 temp;


  Uint8  sheen = chrsheen[character];
  float          spekularity = ((float)sheen / (float)MAXSPEKLEVEL);
  Uint16 model = chrmodel[character];
  Uint16 texture = chrtexture[character];
  Uint16 frame = chrframe[character];
  Uint16 lastframe = chrlastframe[character];
  Uint8 lip = chrlip[character] >> 6;
  Uint8 lightrotation = (chrturnleftright[character] + chrlightturnleftright[character]) >> 8;
  Uint8 ambilevel = chrlightambi[character];
  Uint8 speklevel = chrlightspek[character];



  float uoffset = textureoffset[chruoffset[character] >> 8];
  float voffset = textureoffset[chrvoffset[character] >> 8];
  Uint8 rs = chrredshift[character];
  Uint8 gs = chrgrnshift[character];
  Uint8 bs = chrblushift[character];
  Uint16 lightold, lightnew;

  // Original points with linear interpolation ( lip )
  switch (lip)
  {
  case 0:  // 25% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[frame][cnt] + temp >> 2);

      temp = madvrty[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[frame][cnt] + temp >> 2);

      temp = madvrtz[lastframe][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[frame][cnt] + temp >> 2);

      v[cnt].nx = kMd2Normals[madvrta[frame][cnt]][0]*0.25 + kMd2Normals[madvrta[lastframe][cnt]][0]*0.75;
      v[cnt].ny = -(kMd2Normals[madvrta[frame][cnt]][1]*0.25 + kMd2Normals[madvrta[lastframe][cnt]][1]*0.75);
      v[cnt].nz = -(kMd2Normals[madvrta[frame][cnt]][2]*0.25 + kMd2Normals[madvrta[lastframe][cnt]][2]*0.75);


      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);

      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].a = 1.0f;
    }
    break;

  case 1:  // 50% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float)(madvrtx[frame][cnt] +
                         madvrtx[lastframe][cnt] >> 1);
      v[cnt].y = (float)(madvrty[frame][cnt] +
                         madvrty[lastframe][cnt] >> 1);
      v[cnt].z = (float)(madvrtz[frame][cnt] +
                         madvrtz[lastframe][cnt] >> 1);

      v[cnt].nx = kMd2Normals[madvrta[frame][cnt]][0]*0.5 + kMd2Normals[madvrta[lastframe][cnt]][0]*0.5;
      v[cnt].ny = -(kMd2Normals[madvrta[frame][cnt]][1]*0.5 + kMd2Normals[madvrta[lastframe][cnt]][1]*0.5);
      v[cnt].nz = -(kMd2Normals[madvrta[frame][cnt]][2]*0.5 + kMd2Normals[madvrta[lastframe][cnt]][2]*0.5);

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);

      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].a = 1.0f;
    }
    break;

  case 2:  // 75% this frame
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      temp = madvrtx[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].x = (float)(madvrtx[lastframe][cnt] + temp >> 2);
      temp = madvrty[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].y = (float)(madvrty[lastframe][cnt] + temp >> 2);
      temp = madvrtz[frame][cnt];
      temp = temp + temp + temp;
      v[cnt].z = (float)(madvrtz[lastframe][cnt] + temp >> 2);

      v[cnt].nx = kMd2Normals[madvrta[frame][cnt]][0]*0.75 + kMd2Normals[madvrta[lastframe][cnt]][0]*0.25;
      v[cnt].ny = -(kMd2Normals[madvrta[frame][cnt]][1]*0.75 + kMd2Normals[madvrta[lastframe][cnt]][1]*0.25);
      v[cnt].nz = -(kMd2Normals[madvrta[frame][cnt]][2]*0.75 + kMd2Normals[madvrta[lastframe][cnt]][2]*0.25);

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);

      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].a = 1.0f;
    }
    break;

  case 3:  // 100% this frame...  This is the legible one
    for (cnt = 0; cnt < madtransvertices[model]; cnt++)
    {
      v[cnt].x = (float) madvrtx[frame][cnt];
      v[cnt].y = (float) madvrty[frame][cnt];
      v[cnt].z = (float) madvrtz[frame][cnt];

      v[cnt].nx = kMd2Normals[madvrta[frame][cnt]][0];
      v[cnt].ny = -kMd2Normals[madvrta[frame][cnt]][1];
      v[cnt].nz = -kMd2Normals[madvrta[frame][cnt]][2];

      lightnew = MIN(255, speklevel * spek_local[lightrotation][madvrta[frame][cnt]] + spekularity * 255 * spek_global[lightrotation][madvrta[frame][cnt]]);
      lightnew = lighttospek[sheen][lightnew] + ambilevel + (lightambi*255);
      lightold = chrvrta[character][cnt];
      chrvrta[character][cnt] = MIN(255, 0.9 * lightold + 0.1 * lightnew);

      v[cnt].r = (float)(chrvrta[character][cnt] >> rs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].g = (float)(chrvrta[character][cnt] >> gs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].b = (float)(chrvrta[character][cnt] >> bs) / 255.0f * (float)(trans) / 255.0f;
      v[cnt].a = 1.0f;
    }
    break;

  }

  /*
  // Do fog...
  if(fogon && chrlight[character]==255)
  {
  // The full fog value
  alpha = 0xff000000 | (fogred<<16) | (foggrn<<8) | (fogblu);

  for (cnt = 0; cnt < madtransvertices[model]; cnt++)
  {
  // Figure out the z position of the vertex...  Not totally accurate
  z = (v[cnt].z * chrscale[character]) + chrmatrix[character](3,2);

  // Figure out the fog coloring
  if(z < fogtop)
  {
  if(z < fogbottom)
  {
  v[cnt].specular = alpha;
  }
  else
  {
  spek = v[cnt].specular & 255;
  z = (z - fogbottom)/fogdistance;  // 0.0 to 1.0...  Amount of old to keep
  fogtokeep = 1.0-z;  // 0.0 to 1.0...  Amount of fog to keep
  spek *= z;
  red = (fogred * fogtokeep) + spek;
  grn = (foggrn * fogtokeep) + spek;
  blu = (fogblu * fogtokeep) + spek;
  fogspec = 0xff000000 | (red<<16) | (grn<<8) | (blu);
  v[cnt].specular = fogspec;
  }
  }
  }
  }
  */

  ATTRIB_PUSH("render_texmad", GL_ENABLE_BIT | GL_TRANSFORM_BIT | GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_CURRENT_BIT);
  {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixf(chrmatrix[character].v);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // Choose texture and matrix
    GLTexture_Bind(&TxTexture[texture], CData.texturefilter);

    // Render each command
    entry = 0;
    for (cnt = 0; cnt < madcommands[model]; cnt++)
    {
      glBegin(madcommandtype[model][cnt]);

      for (tnc = 0; tnc < madcommandsize[model][cnt]; tnc++)
      {
        vertex = madcommandvrt[model][entry];

        glColor4fv(&v[vertex].r);
        glTexCoord2f(madcommandu[model][entry] + uoffset, madcommandv[model][entry] + voffset);
        glVertex3fv(&v[vertex].x);
        glNormal3fv(&v[vertex].nx);

        entry++;
      }
      glEnd();
    }

    glPopMatrix();
  }
  ATTRIB_POP("render_texmad");

}

//--------------------------------------------------------------------------------------------
void render_mad(Uint16 character, Uint8 trans)
{
  // ZZ> This function picks the actual function to use
  Sint8 hide = caphidestate[chrmodel[character]];

  if (hide == NOHIDE || hide != chraistate[character])
  {
    if (chrenviro[character])
      render_enviromad(character, trans);
    else
      render_texmad(character, trans);
  }
}

//--------------------------------------------------------------------------------------------
void render_refmad(int tnc, Uint8 trans)
{
  // ZZ> This function draws characters reflected in the floor
  if (capreflect[chrmodel[tnc]])
  {
    int alphatemp;
    int level = chrlevel[tnc];
    int zpos = (chrmatrix[tnc])_CNV(3, 2) - level;
    int model = chrmodel[tnc];
    Uint16 lastframe = chrlastframe[tnc];

    alphatemp = trans - (zpos >> 1);
    if (alphatemp < 0) alphatemp = 0;
    if (alphatemp > 255) alphatemp = 255;

    if (alphatemp > 0)
    {
      Uint8 sheensave = chrsheen[tnc];
      chrredshift[tnc] += 1;
      chrgrnshift[tnc] += 1;
      chrblushift[tnc] += 1;
      chrsheen[tnc] >>= 1;
      (chrmatrix[tnc])_CNV(0, 2) = -(chrmatrix[tnc])_CNV(0, 2);
      (chrmatrix[tnc])_CNV(1, 2) = -(chrmatrix[tnc])_CNV(1, 2);
      (chrmatrix[tnc])_CNV(2, 2) = -(chrmatrix[tnc])_CNV(2, 2);
      (chrmatrix[tnc])_CNV(3, 2) = -(chrmatrix[tnc])_CNV(3, 2) + level + level;
      zpos = fogon;
      fogon = bfalse;

      render_mad(tnc, alphatemp);

      fogon = zpos;
      (chrmatrix[tnc])_CNV(0, 2) = -(chrmatrix[tnc])_CNV(0, 2);
      (chrmatrix[tnc])_CNV(1, 2) = -(chrmatrix[tnc])_CNV(1, 2);
      (chrmatrix[tnc])_CNV(2, 2) = -(chrmatrix[tnc])_CNV(2, 2);
      (chrmatrix[tnc])_CNV(3, 2) = -(chrmatrix[tnc])_CNV(3, 2) + level + level;
      chrsheen[tnc] = sheensave;
      chrredshift[tnc] -= 1;
      chrgrnshift[tnc] -= 1;
      chrblushift[tnc] -= 1;

    }


  }
}

//#if 0
////--------------------------------------------------------------------------------------------
//void render_texmad(Uint16 character, Uint8 trans)
//{
//  Md2Model *model;
//  Uint16 texture;
//  int frame0, frame1;
//  float lerp;
//  GLMATRIX mTempWorld;
//
//  // Grab some basic information about the model
//  model = md2_models[chrmodel[character]];
//  texture = chrtexture[character];
//  frame0 = chrframe[character];
//  frame1 = chrframe[character];
//  lerp = (float)chrlip[character] / 256.0f;
//
//  mTempWorld = mWorld;
//  /*
//  // Lighting information
//  Uint8 lightrotation =
//  (chrturnleftright[character]+chrlightturnleftright[character])>>8;
//  Uint8 speklevel = chrlightspek[character];
//  Uint32 alpha = trans<<24;
//  Uint8 spek = chrsheen[character];
//
//  float uoffset = textureoffset[chruoffset[character]>>8];
//  float voffset = textureoffset[chrvoffset[character]>>8];
//  Uint8 rs = chrredshift[character];
//  Uint8 gs = chrgrnshift[character];
//  Uint8 bs = chrblushift[character];
//
//
//  if(CData.phongon && trans == 255)
//  spek = 0;
//  */
//
//  // Choose texture and matrix
//  if(SDLKEYDOWN(SDLK_F7))
//  {
//    glBindTexture( GL_TEXTURE_2D, -1 );
//  }
//  else
//  {
//    GLTexture_Bind( &TxTexture[texture], CData.texturefilter );
//  }
//
//  mWorld = chrmatrix[character];
//
//  glLoadMatrixf(mView.v);
//  glMultMatrixf(mWorld.v);
//
//  draw_textured_md2(character, model, frame0, frame1, lerp);
//
//
//  mWorld = mTempWorld;
//  glLoadMatrixf(mView.v);
//  glMultMatrixf(mWorld.v);
//}
//#endif
